/*  smp.c: This file main file for the Solarmax proxy.
 *  Version: 0.22
 *  Authors: Axel C. Voigt
 *           Peter Mhlbeyer
 *  Date:    October 26, 2009
 *
 *  The Solarmax proxy is designed to abstract different power inverters for
 *  e.g. the solarview software written by Manfred Richter (manfred@amhamberg.de)
 *  and  maybe other software that supports Solarmax power inverters.
 *
 *  I wrote this piece of software since I have sunways NT4000 power inverters
 *  and they have not been supported by the solarview software. Feel free to
 *  adapt the software to suit your needs. This software comes without warranty.
 *
 *  If you find this software useful, if you can now watch your power inverters
 *  remotely and you cannot program (to help me adapting this software to more
 *  power inverter types, you can contribute to me, even small amounts help :-)
 *
 *  At the moment only NT4000 was tested, testers for other inverters are welcome.
 *    paypal (avau@gmx.de - subject SolarMaxProxy)
 *  or contact me if you could help me developing this piece of software.
 *
 *  history: v0.01 - sunways nt4000 supported, tested with two inverters
 *           v0.12 - mipsel binary, daemonized, arguments
 *           v0.13 - clean up code, reconnect to inverter, connect when needed
 *                   guess correct model number from serial[6]
 *           v0.14 - checks if already running, Sunways errors converted to
 *                   Solarmax in SYS, not tested yet
 *           v0.15 - error codes corrected (as discussed with Manfred Richter)
 *                   kyr (yearly value) is calculated correctly now,
 *                   debug mode implemented, logfile implemented
 *           v0.16 - connect behaviour, error code when inverter not reachable,
 *                   kyr (yearly value) is calculated correctly now (2nd try :-)),
 *                   much more logging (-d necessary),
 *                   pid file generation (-r changes default /var/run/smp),
 *                   father/child processes created (father cares about child)
 *           v0.17 - sig handling improved, restarting of crashing child enabled,
 *                   socket handling comm_connect/comm_disconnect changed,
 *                   checks for NULL pointers added, logging improved
 *           v0.18 - Power setting (cmdline switch -w) introduced. This setting
 *                   overwrites the default of 6950 KWp (my inverters). Currently
 *                   it is not supported to set inverters of different power.
 *           v0.19 - internal release
 *           v0.20 - comm process does not crash when answer timeout from inverter,
 *                   -d switch extended (-d 0 (no debug), -d 1 (std debug), -d 2
 *                   (comm debug)),
 *                   retries sending request to inverter when no answer,
 *                   cuts answer when more then needed bytes to prevent error,
 *                   re-reads when not enough bytes received,
 *                   problem with -t parameter fixed
 *           v0.21 - problem solved, that unsuccessful query (timeout etc.) to
 *                   inverter returns old (cached) data, extended -d switch removed,
 *                   source code see: http://sourceforge.net/projects/solarmax-proxy/
 *           v0.22 - by Peter Mhlbeyer
 *                   Program new structured, comments inserted, start options corrected,
 *                   NT protocol added partly in sw_func.c, replaced pin by pp (peak power),
 *                   added time for debugging after CheckCommand, renamed general.h to smp.h,
 *                   modified simulation mode w/ parabolic power/energy distribution and
 *                   reset of day's energy /kdy) after midnight,
 *                   changed some function's names (setInverterTime, getInverterValues,
 *                   sm_CheckCommand, sm_PrepareAnswer), moved comm_request to sw_func.c
 *                   and renamed to sw_commRequest.
 */

/* standard includes */
#include <stdio.h>
#include <time.h>
#include <sys/wait.h>
#include <sys/types.h> 
#include <signal.h>

/* project includes */
#include "smp.h"
#include "sockets.h"
#include "inv_func.h"
#include "sm_func.h"
#include "comm_func.h"

#define BRAND_SUNWAYS 0      /* NT4000 is tested */
#define BRAND_SMA     1      /* not yet supported */
#define BRAND_KACO    2      /* not yet supported */
#define LISTEN_PORT   12345  /* same as solarview-fb */
#define INVERTER_CNT  1      /* number of inverters */
#define INVERTER_PORT 10000  /* inverter port */
#define INVERTER_PP   6950   /* installed peak power [Wp] */
#define INV_HOST    " 127.0.0.1" /* inverter host */
#define LOGFILE      "/var/log/smp.log"
#define PIDFILE      "/var/run/smp"
#define USAGESTRING  "[-h <inv host>][-p <inv port>][-w pp (Wp)][-c <inv count>][-l <listenport>][-f logfile][-r pidfile][-t timeout][-d][-s][-v]"
#define MAXLINE      (80 + 1)
#define TMPDIR      "/tmp"
#define OPTSTRING    "h:p:w:c:l:f:r:t:dsv"

static char _sccsid[] = { " smp v0.22 2013-12-29 by acv " };
solarmax_data smdata[MAX_UNITS+1];  /* array of values */

char progname[ MAXLINE ];
char logfile[ MAXLINE ];
char pidfile[ MAXLINE ];
FILE* logfp;
char invhost[ MAXLINE ];
char msgline[ MAXLINE + 1 ];
int test_mode;
int debug= 0;
unsigned short gAnsTimeout= 1;
pid_t pid= 0; /* working child */
short run= 1;

/* prototypes */
int list_process_ids (const char *name);
int loop(int sim, char* cmdline_hostname, int cmdline_port, int cmdline_invcount, int listenport, int cmdline_pp);
void SignalHandler(int);

/* ------------------------------------------------------------------------- */
/* init                                                                      */
/* ------------------------------------------------------------------------- */
void init(int sim, int pp)
{
	int z= 0;
	for (z=0;z<MAX_UNITS;z++)
	{
		if(sim==0)
		{
			smdata[z].dyr= 0;
			smdata[z].dmt= 0;
			smdata[z].ddy= 0;
			smdata[z].pac= 0;
			smdata[z].kyr= 0;
			smdata[z].kmt= 0;
			smdata[z].kdy= 0;
			smdata[z].kt0= 0;
			smdata[z].typ= 0;
			smdata[z].prl= 0;
			smdata[z].udc= 0;
			smdata[z].ul1= 0;
			smdata[z].idc= 0;
			smdata[z].il1= 0;
			// smdata[z].tkk= 0;
			// smdata[z].swv= 0;
			// smdata[z].sys= 0;
			smdata[z].pp= pp;
			smdata[z].sal= 0;
		} else
		{
			smdata[z].dyr= 0x9;
			smdata[z].dmt= 0xA;
			smdata[z].ddy= 0x9;
			smdata[z].pac= 0x662;
			smdata[z].kyr= 0x1532;
			smdata[z].kmt= 0x7A;
			smdata[z].kdy= 0;
			smdata[z].kt0= 0x1E2B;
			smdata[z].typ= 0x1201;
			smdata[z].prl= 0xE;
			smdata[z].udc= 0xE69;
			smdata[z].ul1= 0x938;
			smdata[z].idc= 0xEA;
			smdata[z].il1= 0x162;
			// smdata[z].tkk= 0x27;
			// smdata[z].swv= 0xF;
			// smdata[z].sys= 0x4E28,0;
			smdata[z].pp= pp;
			smdata[z].sal= 0;
		}
	}
}
/* ----- init -------------------------------------------------------------- */

/* ------------------------------------------------------------------------- */
/* SearchInverter                                                            */
/* ------------------------------------------------------------------------- */
int SearchInverter(int adr)
{
	int z= 0;
	for(z=0;z<=MAX_UNITS;z++)
	{
		if(smdata[z].adr == adr) return 1;
	}
	return 0;
}
/* ----- SearchInverter ---------------------------------------------------- */

/* ------------------------------------------------------------------------- */
/* waitForConnection                                                         */
/* ------------------------------------------------------------------------- */
Socket* waitForConnection(Socket* skt)
{
	Socket* s= NULL;
	if(debug==1) fprintf(logfp, "smp - waiting for connection ...\n");
	if(debug==1) fflush(logfp);
	s=Saccept(skt);
	if(s)
		if(debug==1) fprintf(logfp, "smp - connected!\n");
	else
		if(debug==1) fprintf(logfp, "smp - ERROR - no connection!\n");
	
	if(debug==1) fflush(logfp);
	return s;
}
/* ----- waitForConnection ------------------------------------------------- */

/* ------------------------------------------------------------------------- */
/* sendError, Solarmax specific                                              */
/* ------------------------------------------------------------------------- */
void sendError(Socket* s, unsigned char* invadr, unsigned char* masteradr)
{
	int cs_calc= 0;
	unsigned char result[255];
	int z= 0;

	sprintf(result, "{%s;%s;07|3E8:IPN|", invadr, masteradr);
	/* calculate checksum */
	for(z=1;z<strlen(result);z++)
	 {
		cs_calc+=result[z];
	}
	sprintf(result, "%s%X}", result, cs_calc);
	Sprintf(s,result);
}
/* ----- sendError --------------------------------------------------------- */

/* ------------------------------------------------------------------------- */
/* SignalHandler                                                             */
/* ------------------------------------------------------------------------- */
void SignalHandler(int sig)
{
	fprintf(logfp, "smp - FATHER caught signal no. %d\n", sig);fflush(logfp);
	signal(sig, SIG_IGN);
	switch(sig)
	{
		case SIGTERM:
		case SIGKILL:
		{
			fprintf(logfp, "smp - FATHER sets run=0 and sigterm CHILD(%d)\n", pid);fflush(logfp);
			run= 0;
			if(pid>0) kill(pid, SIGTERM); /* kill child */
		}
		break;
		default:
		break;
	}
	signal(SIGTERM, SignalHandler);
}
/* ----- SignalHandler ----------------------------------------------------- */

/* ------------------------------------------------------------------------- */
/* loop                                                                      */
/* ------------------------------------------------------------------------- */
int loop(int sim, char* cmdline_hostname, int cmdline_port, int cmdline_invcount, 
     int listenport, int cmdline_pp)
{
	Socket *srvr = NULL;
	Socket *skt  = NULL;
	Socket *inv  = NULL;

	char tmp[255];
//	int len= 0;
	unsigned int z=0;
	int ret=0;
	unsigned short numberOfInverters= 0; /* number of inverters found */

	int cmdline_brand=BRAND_SUNWAYS;
	char port[21];

/* variables for simulation*/
	struct timeval tv;
	long day_sec;
	long old_sec[MAX_UNITS];
	float p_float;
	int p_int;

	/* initialize array */
	init(sim, cmdline_pp);
	setInverterType(cmdline_brand);
	for (z=0; z<MAX_UNITS; z++) { old_sec[z]=86400; }   /* initialize old_sec for midnight */
	z=0;

	if(sim==0)
	{
		if(debug==1) fprintf( logfp,"\nsmp - scanning for %d inverters ...", cmdline_invcount);
		inv= comm_connect(cmdline_hostname, cmdline_port); /* connect */
		if(!inv)
		{
			fprintf( logfp,"smp - ERROR - connecting to inverter%s!\n", cmdline_invcount>1?"s":"");
			fflush(logfp);
		}
		else
		{
			numberOfInverters= scanInverters(inv, cmdline_invcount);
			if(debug==1) fprintf( logfp,"%d found", numberOfInverters);

			for(z=0;z<numberOfInverters;z++)
			{
				if(debug==1) fprintf(logfp, " %s[%2.0d:%s]", getModelStr(smdata[z].typ, tmp), 
					smdata[z].adr, smdata[z].ser);
			}
			if(debug==1) fprintf( logfp,"!\n"); fflush(logfp);
				comm_disconnect(inv); /* disconnect from inverter */
		}
	}
	else { numberOfInverters= 1; } /* simulate one connection */

	sprintf(port, "S%d", listenport);
	srvr= Sopen("", port);
	if(srvr)
	{
		if(debug==1) fprintf( logfp,"smp - ready to receive orders at port %d\n", listenport);
	}
	else
	{
		fprintf( logfp,"smp - ERROR - opening port %d\n", listenport);
		return -1;
	}
	fflush(logfp);

	while(run==1 && (NULL!=(skt=waitForConnection(srvr))))
	{
		char result[2048];
		char invadr[3];
		char masteradr[3];
		char msglen[4];
		char data[255];
		char newValues= -9;

		if(debug==1) fprintf( logfp,"smp - wait for data ...\n");
		if(debug==1) fflush(logfp);

		while(skt&&(0<=Stimeoutwait(skt, 2, 0))) /* wait for data */
		{
			usleep(500*1000);
			if(debug==1) { fprintf( logfp,"smp - checking command ... \n"); fflush(logfp); }
			if(0!=(ret=sm_CheckCommand(skt,
				(char*)&invadr,
				(char*)&masteradr,
				(char*)&msglen,
				(char*)&data)))
			{
			int cs_calc= 0;

			fprintf( logfp,"smp - ERROR - did not understand query!\n");fflush(logfp);

			/* if these data were not correct, set defaults */
			if(ret>=-4)
			{
				strcpy(invadr, "01");
				strcpy(masteradr, "FB");
			}
			/* reply with error */
			sendError(skt, invadr, masteradr);
			continue;
			}
			if(debug==1) { fprintf( logfp,"smp - CheckCommand passed with %d (%s MODE)\n", 
				ret, sim==0?"NORMAL":"SIMULATION"); fflush(logfp); }

			if(sim==0) 
			{
				if(debug==1) {fprintf( logfp,"smp - trying to connect (%s:%d) ... \n", 
					cmdline_hostname, cmdline_port); fflush(logfp);}
				inv= comm_connect(cmdline_hostname, cmdline_port);
				if(!inv)
				{
					int cs_calc= 0;
					fprintf( logfp,"smp - ERROR - connecting to inverter%s at %s:%d!\n", 
						cmdline_invcount>1?"s":"",cmdline_hostname,cmdline_port);
					fflush(logfp);
					comm_disconnect(inv);

					/* reply with error */
					sendError(skt, invadr, masteradr);                                	
					continue;
				}

				if(0==SearchInverter(atoi(invadr)))
				{
					char tmp[10];
					if(debug==1)
					{
						fprintf( logfp,"smp - inverter %s not found! (re)scan ... ", invadr);
						fflush(logfp);
					}

					numberOfInverters= scanInverters(inv, cmdline_invcount);
					if(debug==1)
					{
						fprintf( logfp,"%d found", numberOfInverters);
						fflush(logfp);
					}

					for(z=0;z<numberOfInverters;z++)
					{
						if(debug==1) fprintf(logfp, " %s[%2.0d:%s]", getModelStr(smdata[z].typ, tmp), 
							smdata[z].adr, smdata[z].ser);
					}
					if(debug==1) {fprintf( logfp,"!\n");fflush(logfp);}

					/* inverter not found in local list */
					if( 0==SearchInverter(atoi(invadr)) )
					{
						fprintf( logfp,"smp - ERROR - inverter %s not found or could not be contacted!\n", invadr);
						fflush(logfp);
						comm_disconnect(inv);

						/* reply with error */
						sendError(skt, invadr, masteradr);
						continue;
					}
				}
			}

			/* new values? */
			if( sim==0 )
			{
				if(debug==1) {fprintf( logfp,"smp - querying inverter %s\n", invadr);fflush(logfp);}
				newValues= getInverterValues(inv,atoi(invadr));
				if(debug==1) {fprintf( logfp,"smp - querying inverter %s returned %d\n", 
					invadr, newValues);fflush(logfp);}
			}
			else
			{
				newValues=0; /* simulation */

				// smdata[atoi(invadr)-1].pac+=1000; /* add 10 Watt */
				// smdata[atoi(invadr)-1].kdy+=0.05; /* add 0.05 kWh */
				// smdata[atoi(invadr)-1].kt0+=0.05; /* add 0.05 kWh */

				/* get time since 01.01.1970, 1 h difference to actual value ??? */
				gettimeofday (&tv, NULL);
				day_sec = tv.tv_sec % 86400;        /* calculate today's seconds with modulo */
				/* calculate parabola of power acc. formula p = -a*(t - c)^2 + b             */
				/* where a = 3,858E-6, b = 800 [W], c = 43200 [s], p = power, t = time [s]   */
				p_float = (day_sec - 43200)*(day_sec - 43200)*(-3.858)/1000000 + 800;
				if (p_float <= 0) p_float = 0;      /* no negative value */

				/* reset day's energy values during day shift for all channels */
				if (day_sec < old_sec[atoi(invadr)-1]) { smdata[atoi(invadr)-1].kdy = 0; }

				old_sec[atoi(invadr)-1] = day_sec;

				p_int = (int)p_float*100;
				p_float = p_float*20/3600000;            /* now energy [kWh], assumes solarview poll every 20 s */
				if(debug==1) {fprintf( logfp,"smp - simulation: time %i s, power %i W, energy %f kWh\n", 
					day_sec, p_int/100, p_float);fflush(logfp);}
				smdata[atoi(invadr)-1].pac=p_int;       /* power, should be parabolic */
				smdata[atoi(invadr)-1].kdy+=p_float;    /* add energy */
				smdata[atoi(invadr)-1].kt0+=p_float;    /* add energy */
			}

			if( 0==newValues )
			{
				sm_PrepareAnswer(atoi(invadr), invadr, masteradr, (solarmax_data*)&smdata, result);
				if(debug==1) fprintf( logfp,"smp - result >%s<\n", result);
				Sprintf(skt,result);
			}
			else
			{
				fprintf(logfp, "smp - ERROR - getting values (err:%d)\n", newValues);
				fflush(logfp);
				sendError(skt, invadr, masteradr);
			}
			comm_disconnect(inv);
		}
		Sclose(skt);

		if( 0==newValues )
		{
			/* do total calculations */
			smdata[MAX_UNITS].pac= 0;
			smdata[MAX_UNITS].kdy= 0;
			smdata[MAX_UNITS].kt0= 0;
			for (z=0;z<numberOfInverters;z++)
			{
				smdata[MAX_UNITS].pac+= smdata[z].pac;
				smdata[MAX_UNITS].kdy+= smdata[z].kdy;
				smdata[MAX_UNITS].kt0+= smdata[z].kt0;
			}
			if(debug==1) fprintf( logfp,"\nsmp - WMom: %0.2fW ", smdata[MAX_UNITS].pac/100);
			if(debug==1) fprintf( logfp,"WhDay: %.2fkWh (%0.2f EUR) ", smdata[MAX_UNITS].kdy,
				(double)smdata[MAX_UNITS].kdy*0.431 );
			if(debug==1) fprintf( logfp,"kWhAll: %.1fkWh  (%0.2f EUR)\n\n", smdata[MAX_UNITS].kt0, 
				(smdata[MAX_UNITS].kt0)*0.431 );
		}
		fflush(logfp);
	}

		Sclose(skt);
		Sclose(srvr);

	fprintf( logfp,"smp - terminating ...\n");
	fclose(logfp);
	return 0;
}
/* ----- loop -------------------------------------------------------------- */

/* ------------------------------------------------------------------------- */
/* main                                                                      */
/* ------------------------------------------------------------------------- */
int main( argc, argv )
int argc;
char *argv[];
{
	extern int getopt();
	extern int optind;
	extern char *optarg; /* captures arguments to options */

	int status = 0;
	int c_opt;
	int listenport, invcount, invport, invpp;
	/* end declarations */

	/* check if running */
	if(0<list_process_ids ("smp"))
	{
		printf("smp - process is already running ...\n");
		exit(-1);
	}

	/* inits */
	strcpy( progname, argv[ 0 ] );
	strcpy( logfile, LOGFILE );
	strcpy( pidfile, PIDFILE );
	strcpy(invhost, INV_HOST);
	test_mode = 0;
	listenport= LISTEN_PORT;
	invcount=   INVERTER_CNT;
	invport=    INVERTER_PORT;
	invpp=      INVERTER_PP;

	while( ( c_opt = getopt( argc, argv, OPTSTRING ) ) != EOF )
	{
		switch( c_opt )
		{
			case 'v':
				/* -v (just print version and exit) */
				printf( "%s\n", _sccsid );
				exit( 0 );
			break;
			case 's':
				/* -s (set test mode) */
				test_mode = 1;
			break;
			case 'd':
				/* -d (set debug mode) */
				debug = 1;
			break;
			case 'f':
				/* -f arg */
				if( sscanf( optarg, "%s", logfile ) != 1 )
				{
					fprintf( stderr, "%s: unreadable log file argument\n", progname );
					exit( 1 );
				}
			break;
			case 'r':
				/* -r arg */
				if( sscanf( optarg, "%s", pidfile ) != 1 )
				{
					fprintf( stderr,"%s: unreadable pid file argument\n", progname );
					exit( 1 );
				}
			break;
			case 'c':
				/* -c arg */
				if( sscanf( optarg, "%d", &invcount ) != 1 )
				{
					fprintf( stderr, "%s: unreadable inverter count argument\n", progname );
					exit( 1 );
				}
			break;
			case 'l':
				/* -l arg */
				if( sscanf( optarg, "%d", &listenport) != 1 )
				{
					fprintf( stderr, "%s: unreadable listen port argument\n", progname );
					exit( 1 );
				}
			break;
			case 'p':
				/* -p arg */
				if( sscanf( optarg, "%d", &invport ) != 1 )
				{
					fprintf( stderr, "%s: unreadable inverter port argument\n", progname );
					exit( 1 );
				}
			break;
			case 'w':
				/* -w arg */
					if( sscanf( optarg, "%d", &invpp ) != 1 )
					{
						fprintf( stderr, "%s: unreadable inverter peak power argument\n", progname );
						exit( 1 );
					}
			break;
			case 'h':
				/* -h arg */
				if( sscanf( optarg, "%s", invhost ) != 1 )
				{
					fprintf( stderr, "%s: unreadable inverter host argument\n", progname );
					exit( 1 );
				}
			break;
			case 't':
				/* -t timeout for answers from inverter */
				if( sscanf( optarg, "%d", &gAnsTimeout ) != 1 )
				{
					fprintf( stderr, "%s: unreadable inverter timeout argument, setting timeout to %d\n", 
						progname, gAnsTimeout );
				}
			break;
			default:
				fprintf( stderr, "%s: bad option -%c\n", argv[ 0 ], c_opt );
				fprintf( stderr, "usage: %s %s\n", argv[ 0 ], USAGESTRING );
				exit( 1 );
		}
	}

	printf("************************************************************\n");
	printf("*      Solarmax proxy (%s)      *\n", _sccsid);
	printf("*                                                          *\n");
	if(test_mode==1)
	{
		printf("*   !!!SIMULATION!!! not connecting to any inverters !!!   *\n");
		printf("*                                                          *\n");
	}
	printf("*  smp [-h <host>] [-p <port>] [-l <listenport>]           *\n");
	printf("       [-w <peak power> ] [-c <count>] [-t <timeout>]      *\n");
	printf("*      [-f <logfile>] [-r <pidfile] [-d] [-s] [-v]         *\n");
	printf("*                                                          *\n");
	printf("*  host:        ip/name of inverter or ser2ip adapter      *\n");
	printf("*  port:        port of inverter or ser2ip adapter         *\n");
	printf("*  listenport:  portnumber where smp listens for requests  *\n");
	printf("*  peak power:  installed peak power in Wp                 *\n");
	printf("*  count:       number of inverters                        *\n");
	printf("*  timeout:     timeout in s for inverter answers          *\n");
	printf("*  logfile:     file to log to                             *\n");
	printf("*  pidfile:     file for process ids                       *\n");
	printf("*  -d           debug mode (many more logging)             *\n");
	printf("*  -s           simulation mode (does not query inverter)  *\n");
	printf("*  -v           print version and exit                     *\n");
	printf("************************************************************\n\n");
	printf("watching %d %s at %s:%d listenport: %d (debug=%d, timeout=%d s)\n",
		invcount, "SUNWAYS", invhost, invport, listenport, debug, gAnsTimeout );

	logfp= fopen(logfile, "a");
	if(!logfp)
	{
		logfp= stderr;
		fprintf( stderr,"\nsmp - ERROR - cannot open logfile >%s<, using stderr!\n", logfile);fflush(logfp);
	}

	/* on exiting loop of option parsing optind indexes the next
	 * argv[] argument to the function.
	 */
	if( signal( SIGINT, SIG_IGN ) != SIG_IGN )
		signal( SIGINT, SIG_IGN );
	if( signal( SIGKILL, SIG_IGN ) != SIG_IGN )
		signal( SIGKILL, SIG_IGN );

	if(0==fork())
	{
		run= 1;
		while (run==1)
		{
			pid_t p= 0;
			if( (p=fork()) == 0 )
			{
				/* fork a child process, sending the daemon into the background loop. */

				fclose( stdin );
				fclose( stdout );

				signal(SIGTERM, SIG_DFL);

				if(debug==1) {fprintf(logfp, "smp - CHILD running as pid %d\n", getpid());fflush(logfp);}
				loop( test_mode, invhost, invport, invcount, listenport, invpp );
				exit(0);
			}
			else
			{
				int err= 0;
				pid= p; /* set current child globally */
				FILE* f= fopen(pidfile, "w");
				if(!f) {
					fprintf( logfp, "\nERROR cannot open pidfile >%s<\n", pidfile);fflush(logfp);
				}
				else
				{
					fprintf( f, "%d", getpid() );
					fclose(f);
				}
				signal(SIGTERM, SignalHandler);

				/* exit parent process 
				#if( signal( SIGINT, SIG_DFL ) != SIG_DFL )
				#	signal( SIGINT, SIG_DFL );
				#if( signal( SIGKILL, SIG_DFL ) != SIG_DFL )
				#	signal( SIGKILL, SIG_DFL );
				*/

				if(debug==1) {fprintf(logfp, "smp - FATHER waiting for CHILD to die ...\n");fflush(logfp);}
				err= wait( &status );

				/* PeMue: commented out parts have been removed */

				unlink(pidfile);
				signal(SIGTERM, SIG_DFL);
			}
		}
		if(debug==1) {fprintf(logfp,"smp - terminating as requested\n");fflush(logfp);}
	} else exit(0); /* end if( fork() ) */

	/* NOT REACHED */

	exit( 0 );

	return( 0 );

} /* end main */
/* ----- main -------------------------------------------------------------- */