/* sw_func.c: This file contains the Sunways functions for the Solarmax proxy.
 * Version: 0.1
 * Authors: Axel C. Voigt
 * Date: October 26, 2009
 */

/* Sunways protocol from rev. 1.67:
 *
 * serial/RS485 9600 baud, 8/1 bit, no parity
 *
 * to inverter: 5 bytes
 * STX AD KN BN CS
 *
 * STX  start of transmission  0
 * AD   inverter's address     1 ... 99, 255
 * KN   command number         not zero
 * BN   block number           read different blocks from inverter
 * CS   checksum               STX+AD+KN+BN
 *
 * KN read commands
 *      1    failure data      1.0		$rError   = "\x00\x01\x01\x01"; 	WORKING BUT UNCONFIRMED
 *      2    online values     1.0		$rOnline1 = "\x00\x01\x02\x01"; 	
 *      3    day's values      1.0		$rMon1    = "\x00\x01\x03\x01"; 
 *      4    month's values    1.0		$rYear1   = "\x00\x01\x04\x01"; 	
 *      5    min. val. 1st bl. 1.0		nicht in fhem
 *      6    date and time     1.0		$rTime    = "\x00\x01\x06\x01"; 	WORKING
 *      7    min. val. 2nd bl. 1.0		nicht in fhem
 *      8    serial number     1.1		$rSerial  = "\x00\x01\x08\x01";  	WORKING BUT UNCONFIRMED
 *      9    firmware version  1.1		$rProFW   = "\x00\x01\x09\x01"; 	WORKING BUT UNCONFIRMED
 *      10-15min. values NT10000 1.2
 *      15   parameter         1.2
 *      17   display language  1.2
 *
 * KN write commands
 *      50-54set date          1.0
 *      55-57set inv. address  1.0
 *      58   set display lang. 1.2
 *      100-155 set total kWh  1.1
 *      156-175 write parameter1.2
 *
 * from inverter: 13 bytes
 *
 * B1 B2 B3 B4 B5 b6 B7 B8 B9 B10 B11 B12 CS
 * CS = B1 + B2 + ... + B12
 * if Byte Bx not set, then it will be set to 1
 
#-- Process data 
    my $udc = ord($invBuffer[0])*2.8+100;  
    my $idc = ord($invBuffer[1])*0.08;        
    my $uac = ord($invBuffer[2])+100.0;  
    my $iac = ord($invBuffer[3])*0.120;       
    my $t   = ord($invBuffer[4])-40.0;
    my $s   = ord($invBuffer[5])*6.0;
    my $pdc = int($udc*$idc)/1000;
    my $pac = int($uac*$iac)/1000; 
    my $wd  = (ord($invBuffer[6])* 256 + ord($invBuffer[7]))/1000;         
    my $wtot= ord($invBuffer[8])* 256 + ord($invBuffer[9]); 



*/

#include "sw_func.h"
#include "comm_func.h"

extern solarmax_data smdata[MAX_UNITS];
extern int debug;
extern FILE* logfp;
extern unsigned short gAnsTimeout;

/* ------------------------------------------------------------------------- */
/* sw_checkType                                                              */
/* ------------------------------------------------------------------------- */
int sw_checkType( char* serial, char* typname )
{
	/* check typ */
	switch(serial[5])
	{
		case '0':
			strcpy(typname, "Sunways NT10000");
			return NT10000;
		break;
		case '2':
			strcpy(typname, "Sunways NT2600"); 
			return NT2600;
		break;
		case '4':
			strcpy(typname, "Sunways NT4000");
			return NT4000;
		break;
		case '5':
			strcpy(typname, "Sunways NT5000");
			return NT5000;
		break;
		case '6':
			strcpy(typname, "Sunways NT6000");
			return NT6000;
		break;
		default: return -1;
	}
}
/* ----- sw_checkType ------------------------------------------------------ */

/* ------------------------------------------------------------------------- */
/* sw_getModelStr                                                            */
/* ------------------------------------------------------------------------- */
char* sw_getModelStr( int model, char* typstr )
{
	switch(model) {
		case NT2600:
			strcpy(typstr, "NT2600");
		break;
		case NT4000:
			strcpy(typstr, "NT4000");
		break;
		case NT5000:
			strcpy(typstr, "NT5000");
		break;
		case NT6000:
			strcpy(typstr, "NT6000");
		break;
		case NT10000:
			strcpy(typstr, "NT10000");
		break;
		default:
			strcpy(typstr, "not supported");
		break;
	}
	return typstr;
}
/* ----- sw_getModelStr ---------------------------------------------------- */

/* ------------------------------------------------------------------------- */
/* sw_commRequest                                                            */
/* ------------------------------------------------------------------------- */
int sw_commRequest(Socket* skt, short adr, short cmd, short blck, unsigned char* ans, short timeout)
{
	int l= 0;
	int z= 0;
	unsigned char cs= 0;
	char command[5];
	char log[3000];
	struct timeval tv;
	unsigned short retries= 2;
	unsigned short neededBytes= 13; /* std command has 13 bytes answer */

	if( cmd==SW_READ_SERNUM ) /* SW_READ_SERNUM has 24 bytes answer */
	{
		neededBytes= 24;
	}

	command[0]=0;
	command[1]=(char)adr;
	command[2]=(char)cmd;
	command[3]=blck;
	command[4]=command[0]+command[1]+command[2]+command[3];

	strncpy(log, command, 5);
	log[5]='\0';

	while(retries>0)
	{
		l= Swrite(skt, command, 5);
		if(debug==1) 
		{
			for( z=0 ; z<l; z++ )
			{
				sprintf(log, "%s %04X", log, command[z]);
			}
		gettimeofday (&tv, NULL);
		fprintf(logfp,"%ld.%d smp send: >%s< %s (%d bytes)\n", tv.tv_sec, tv.tv_usec, log, l==5?"OK":"ERR", l);
		log[0]='\0';
		}
		if(l<=0)
		return l;

		z= Stimeoutwait(skt, timeout, 0);
		if (z<0)
		{
			retries--;
			gettimeofday (&tv, NULL);
			fprintf(logfp,"%ld.%d ERR TIMEOUT, NO ANSWER! (timeout:%d, retries:%d, code:%d)\n", tv.tv_sec, 
				tv.tv_usec, timeout, retries, z);
				continue;
		}

		retries= 0; /* prevent another loop */
		z= 0;
		while( l>0 && z<neededBytes )
		{
			l= Stimeoutwait(skt, 1, 0); /* wait max 1s for data again */
			if(l<0)
			{
				gettimeofday (&tv, NULL);
				fprintf(logfp,"%ld.%d ERR TIMEOUT, NO MORE DATA!\n", tv.tv_sec, tv.tv_usec);
				break;
			}
			z+= Sread(skt, ans, 255);
			gettimeofday (&tv, NULL);
			if(debug==1)
			{
				fprintf(logfp,"%ld.%d RECV IN LOOP: total %d bytes\n", tv.tv_sec, tv.tv_usec, z);
				fflush(logfp);
			}
		}

		if(debug==1)
		{
			log[0]='\0';
			for( l=0; l<z; l++ )
			{
				sprintf(log, "%s %04X", log, command[l]);
			}
				gettimeofday (&tv, NULL);
				fprintf(logfp,"%ld.%d smp receive: >%s< %s (%d bytes)\n", tv.tv_sec, tv.tv_usec, log, 
					z==neededBytes?"OK":"ERR", z);
		}

		if (z!=neededBytes)
		{
			gettimeofday (&tv, NULL);
			if(z<neededBytes)
			{
				fprintf(logfp,"%ld.%d smp receive: - ERROR - not enough data, aborting (rec:%d,required:%d)\n", 
					tv.tv_sec, tv.tv_usec, z, neededBytes);
				return -2; /* out if not enough data */
			}
			else
			{
				fprintf(logfp,"%ld.%d smp receive: - ERROR - too much data, CUTTING to %d bytes and continue\n", 
					tv.tv_sec, tv.tv_usec, neededBytes);
				ans[neededBytes]='\0';
				z = neededBytes;
			}
		}
	}

	/*if(cmd!=SW_READ_SERNUM) // won't work wth SERIAL 
	{ // calculate checksum 
		for( z=0 ; z<l-1; z++ )
		{
			cs+=ans[z];
		}
		if( cs != ans[l-1] )
		{
			return -3;
		}
	}*/

	gettimeofday (&tv, NULL);
	fprintf(logfp,"%ld.%d smp - receive processed %d bytes\n", tv.tv_sec, tv.tv_usec, z);
	return z;
}
/* ----- sw_commRequest ---------------------------------------------------- */

/* ------------------------------------------------------------------------- */
/* sw_scanInverters                                                          */
/* ------------------------------------------------------------------------- */
int sw_scanInverters(Socket* skt, int max)
{
	int found= 0;
	short address= 1;
	char ans[255];

	if(!skt) return -1;
	while( address<=max )
	{
		/* ask for serial */
		if(0<sw_commRequest(skt, address, SW_READ_SERNUM, 1, ans, gAnsTimeout))
		{
			char typstr[20];
			ans[12]='\0';

			smdata[found].adr= address;
			strncpy(smdata[found].ser, ans, 13);

			smdata[found].typ= sw_checkType( smdata[found].ser, typstr );
			if(smdata[found].typ==-1)
			{
				fprintf(logfp,"smp - ERROR - protocol version < 1.1?\n");
				/* error, could not determine WR type, old version 750V? setting NT2600 */
				smdata[found].typ= NT2600; 
		}
		else if(smdata[found].typ==NT10000)
		{
			fprintf(logfp,"smp - ERROR - NT10000 is not fully supported yet!\n");
		}

		found++;
		address++;
		continue;
	}
	address++;
	}
	return found;
}
/* ----- sw_scanInverters -------------------------------------------------- */

/* ------------------------------------------------------------------------- */
/* sw_setTime,                                                               */
/* ------------------------------------------------------------------------- */
int sw_setTime(Socket* skt)
{
	/* dummy function */
	return 0;
}
/* ----- sw_scanInverters -------------------------------------------------- */

/* ------------------------------------------------------------------------- */
/* sw_getOnlineResults                                                       */
/* ------------------------------------------------------------------------- */
int sw_getOnlineResults(Socket* skt, int num)
{
	int z= 0;
	unsigned char ans[13];
//	double kyr= 0;
	double kmt[13];
	int ret= 0;

	for(z=0;z<13;z++) kmt[z]= 0;

	num--;

	if(0<(ret=sw_commRequest(skt, smdata[num].adr, SW_READ_TIME, 1, ans, gAnsTimeout)))
	{
		smdata[num].ddy= ans[2];
		smdata[num].dmt= ans[1];
		smdata[num].dyr= ans[0];
		smdata[num].thr= ans[3];
		smdata[num].tmi= ans[4];

		if(debug==1) fprintf(logfp,"smp - %d.%d.%d %2d:%2d ", smdata[num].ddy,
			smdata[num].dmt,
			2000+smdata[num].dyr,
			smdata[num].thr,
			smdata[num].tmi);
	}
	else
	{
		fprintf(logfp,"smp - ERROR - getting time from inverter %i (code:%d)!\n", smdata[num].adr, ret);
		fflush(logfp);
		return ret;
	}

	if(0<(ret=sw_commRequest(skt, smdata[num].adr, SW_READ_MONTOT, 1, ans, gAnsTimeout)))
	{
		smdata[num].kmt= (double)(ans[2]<<8|ans[3])/10;

		kmt[0]= (double)(ans[2]<<8|ans[3])/10;
		kmt[1]= (double)(ans[4]<<8|ans[5])/10;
		kmt[2]= (double)(ans[6]<<8|ans[7])/10;
		kmt[3]= (double)(ans[8]<<8|ans[9])/10;
		kmt[4]= (double)(ans[10]<<8|ans[11])/10;
	}
	else
	{
		fprintf(logfp,"smp - ERROR - getting month totals 1st time from inverter %i (code:%d)!\n", 
			smdata[num].adr, ret);
		fflush(logfp);
		return ret;
	}

	if(0<(ret=sw_commRequest(skt, smdata[num].adr, SW_READ_MONTOT, 2, ans, gAnsTimeout)))
	{
		kmt[5]= (double)(ans[0]<<8|ans[1])/10;
		kmt[6]= (double)(ans[2]<<8|ans[3])/10;
		kmt[7]= (double)(ans[4]<<8|ans[5])/10;
		kmt[8]= (double)(ans[6]<<8|ans[7])/10;
		kmt[9]= (double)(ans[8]<<8|ans[9])/10;
		kmt[10]= (double)(ans[10]<<8|ans[11])/10;
	} 
	else
	{
		fprintf(logfp,"smp - ERROR - getting month totals 2nd time from inverter %i (code:%d)!\n", 
			smdata[num].adr, ret);
		fflush(logfp);
		return ret;
	}

	if(0<(ret=sw_commRequest(skt, smdata[num].adr, SW_READ_MONTOT, 3, ans, gAnsTimeout)))
	{
		unsigned long code= ans[9] & 0xC5;
		kmt[11]= (double)(ans[0]<<8|ans[1])/10;
		kmt[12]= (double)(ans[2]<<8|ans[3])/10;
		if(debug==1) fprintf(logfp, "smp - last 12 months: %.2f KWh\n", ((double)(ans[10]<<8|ans[11])+ 0x10000 
			* (code>>6|(code&1)<<2|(code&4)<<1)));
	}
	else
	{
		fprintf(logfp,"smp - ERROR - getting month totals 3rd time from inverter %i (code:%d)\n", 
			smdata[num].adr, ret);
		fflush(logfp);
		return ret;
	}

	/* calculate year */
	smdata[num].kyr= 0;
	if(debug==1) fprintf(logfp, "smp - current month %d\n", smdata[num].dmt);
	for(z=0;z<smdata[num].dmt;z++)
	{
		if(debug==1) fprintf(logfp, "smp - add month %d: %.2f (%.2f)\n", z, kmt[z], smdata[num].kyr);
		smdata[num].kyr+=kmt[z];
	}
	if(debug==1) fprintf(logfp,"smp - WYear:%.1fkWh\n", smdata[num].kyr);

	if(0<(ret=sw_commRequest(skt, smdata[num].adr, SW_READ_ONLINE, 1, ans, gAnsTimeout)))
	{
		smdata[num].udc= (int)((ans[0]*2.8+100)*100); // Udc
		if(debug==1) fprintf(logfp,"smp - UDC:%.2fV ", (double)smdata[num].udc/100);
		smdata[num].idc= ans[1]*0.08*100; // Idc
		if(debug==1) fprintf(logfp,"IDC:%.2fA ", smdata[num].idc/100);
		smdata[num].ul1= (ans[2]+100)*100;	// Uac
		if(debug==1) fprintf(logfp,"UAC:%.2fV ", smdata[num].ul1/100);
		smdata[num].il1= ans[3]*0.12*100; // Iac
		if(debug==1) fprintf(logfp,"IAC:%.2fA ", smdata[num].il1/100);
		smdata[num].pac= (ans[2]+100)*(ans[3]*0.12)*100;
		if(debug==1) fprintf(logfp,"PAC:%0.2fW ", smdata[num].pac/100);
		smdata[num].kdy= (double)(ans[6]<<8|ans[7])/1000; // WTag
		if(debug==1) fprintf(logfp,"WTAG:%.2fkWh ", smdata[num].kdy);
		smdata[num].err= ans[10]; // errcode
		if(debug==1) fprintf(logfp,"ERR:%s ", (smdata[num].err==0||smdata[num].err==128)?"OK":"ERROR");
		smdata[num].ver= ans[11] & 0x3A; // art
		if(debug==1) fprintf(logfp,"VER:%s ", smdata[num].ver==26?"NT850V":"n.a.");
		smdata[num].sal= ans[11] & 0xC5; // code2
		/* erweitert den Gesamtbereich WGES auf 1.048.575 kWh */
		//fprintf(logfp,"BERLAUF:%d ", smdata[num].sal);
		smdata[num].kt0= ((double)(ans[8]<<8|ans[9])) + 0x10000 * (smdata[num].sal>>6|
			(smdata[num].sal&1)<<2|(smdata[num].sal&4)<<1); // WTotal
		if(debug==1) fprintf(logfp,"WGES:%.1fkWh\n", smdata[num].kt0);
		smdata[num].sys= sw_translateStatusCode(smdata[num].err);
	}
	else return ret; // tbd. hier fehlt noch das error logging

	return 0; /* new values available */
}
/* ----- sw_getOnlineResults ----------------------------------------------- */

/* ------------------------------------------------------------------------- */
/* sw_translateStatusCode                                                    */
/* ------------------------------------------------------------------------- */
int sw_translateStatusCode(int status) 
{
	switch (status)
	{
		//case 15: return Offl; /* offline */
		case 16: return 0x4E8F; /* overvoltage - DC */
		case 17: return 0x4E95; /* mains frequency 4E94: too high, 4E95 too low */
		case 18: return 0x4E9C; /* temperature */
		case 19: return 0x4E9B; /* mains supply too low 1st */
		case 20: return 0x4E9A; /* mains supply too high 1st */
		case 21: return 0x4ECD; /* mains supply too low 2nd */
		case 22: return 0x4E92; /* Sunways AFI, error current AC/DC */ 
		case 23: return 0x4E89; /* insulation */
		case 24: return 0x4EE7; /* unknown failure */
		case 25: return 0x4E96; /* "Inselbetrieb" */
		case 26: return 0x4ECC; /* mains supply too high 2nd */
		case 27: return 0x4EC9; /* surge fault, bad mains quality */
		case 28: return 0x4E8F; /* overvoltage - AC */
		case 29: return 0x4EE7; /* unknown failure */

		case 0:  /* no error */
		case 128:/* no error */
		default:
			return 0x4E21; /* 4e21 agreed with Manfred Richter*/
		break;
	}
}
/* ----- sw_translateStatusCode -------------------------------------------- */
